#include "otc/write_dot.h"
#include <algorithm>
#include "otc/embedded_tree.h"
#include "otc/node_embedding.h"
#include "otc/tree.h"
#include "otc/tree_data.h"
#include "otc/tree_iter.h"
#include "otc/tree_operations.h"
#include "otc/forest.h"

namespace otc {
typedef std::pair<const NodeWithSplits *, std::string> ToDotKey;
typedef std::pair<std::string, std::string> NamePair; // nodeName and nodeLabel
typedef std::map<ToDotKey, NamePair> NodeToDotNames;
void writeNodeDOT(std::ostream & out,
                           const ToDotKey & k,
                           NodeToDotNames & nd2name,
                           const std::string &style, 
                           bool forceMRCANaming,
                           bool writeLabel,
                           bool pt);
void uncheckedWriteNodeDOT(std::ostream & out,
                           const ToDotKey & k,
                           NodeToDotNames & nd2name,
                           const std::string &style, 
                           bool forceMRCANaming,
                           bool writeLabel,
                           bool pt);
void writeDOTEdge(std::ostream & out,
                  const ToDotKey anc,
                  const ToDotKey des,
                  NodeToDotNames & nd2name,
                  const std::string &style, 
                  bool toMidEdge);
void writeDOTCrossEdge(std::ostream & out,
                  const ToDotKey anc,
                  const ToDotKey des,
                  NodeToDotNames & nd2name,
                  const std::string &style);
void writePathPairingToDOT(std::ostream & out,
                           const NodeWithSplits *n, 
                           const PathPairingWithSplits & pp,
                           NodeToDotNames & nd2name,
                           std::set<const PathPairingWithSplits *> & pathSet,
                           const char * color,
                           const char * prefix) ;
void writeDOTEmbeddingForNode(std::ostream & out,
                              const NodeWithSplits *n, 
                              const NodeEmbeddingWithSplits & thr,
                              const std::map<const NodeWithSplits *, NodeEmbeddingWithSplits> & eForNd,
                              NodeToDotNames & nd2name,
                              std::set<const PathPairingWithSplits *> & pathSet,
                              const char * color,
                              const char * prefix,
                              std::size_t treeIndex
                              );
void writeOneSideOfPathPairingToDOT(std::ostream & out,
                                     const NodeWithSplits * pn,
                                     const NodeWithSplits * pd,
                                     const NodeWithSplits * sn,
                                     const NodeWithSplits * sd,
                                     const ToDotKey & midPointKey,
                                     const NamePair & namePair,
                                     NodeToDotNames & nd2name,
                                     const NodeWithSplits * focalNode,
                                     const std::string & style,
                                     const char * prefix);
void writeDOTForNodeWithoutEmbedding(std::ostream & out, const NodeWithSplits *nd , NodeToDotNames & nd2name);
std::string edgeStyleForColor(const char * color);
ToDotKey findAnyKey(const NodeWithSplits *nd, const NodeToDotNames & nd2name);
void writeDOTForFtree(std::ostream & out,
                              const FTree<RTSplits, MappedWithSplitsData> & tree, 
                              NodeToDotNames & nd2name,
                              const char * color,
                              std::size_t treeIndex
                              );
// IMPL
std::string edgeStyleForColor(const char * color) {
    std::string bstyle = "fontcolor=\"";
    bstyle.append(color);
    bstyle.append("\" color=\"");
    bstyle.append(color);
    bstyle.append("\"");
    return bstyle;
}
constexpr const char * COLORS [] = {"crimson",
                                    "blue",
                                    "springgreen",
                                    "magenta",
                                    "darkorange",
                                    "lightblue",
                                    "goldenrod",
                                    "brown",
                                    "gray"};
constexpr unsigned LAST_COLOR_IND = 8;

void writeNodeDOT(std::ostream & out,
                           const ToDotKey & k,
                           NodeToDotNames & nd2name,
                           const std::string &style, 
                           bool forceMRCANaming,
                           bool writeLabel,
                           bool pt) {
    if (!contains(nd2name, k)) {
        uncheckedWriteNodeDOT(out, k, nd2name, style, forceMRCANaming, writeLabel, pt);
    }
}

void writeDOTForNodeWithoutEmbedding(std::ostream & , const NodeWithSplits * , NodeToDotNames & ) {
}

void uncheckedWriteNodeDOT(std::ostream & out,
                           const ToDotKey & k,
                           NodeToDotNames & nd2name,
                           const std::string &style, 
                           bool forceMRCANaming,
                           bool writeLabel,
                           bool pt) {
    const NodeWithSplits * nd{k.first};
    const std::string & prefix{k.second};
    std::string name{prefix};
    const std::string unadorned = forceMRCANaming ? get_mrca_designator(*nd) :  get_designator(*nd);
    name.append(unadorned);
    nd2name.emplace(k, NamePair{name, unadorned});
    assert(contains(nd2name, k));
    const NamePair & np = nd2name.at(k);
    out << "  \"" << np.first << "\"[";
    if (pt) {
        out << "shape=point ";
    }
    if ((!pt) && writeLabel) {
        out << "label=\"" << np.second << "\" ";
    } else {
        out << "label=\"\" ";
    }
    if (!style.empty()) {
        out << style;
    }
    out << "];\n";
}

void writeDOTEdge(std::ostream & out,
                  const ToDotKey anc,
                  const ToDotKey des,
                  NodeToDotNames & nd2name,
                  const std::string &style, 
                  bool /*toMidEdge*/) {
    assert(contains(nd2name, anc));
    assert(contains(nd2name, des));
    out << "    \"" << nd2name.at(anc).first << "\" -> \"" << nd2name.at(des).first << '\"';
    if (!style.empty()) {
        out << "[" << style <<']';
    }
    out << ";\n";
}

void writeDOTCrossEdge(std::ostream & out,
                  const ToDotKey anc,
                  const ToDotKey des,
                  NodeToDotNames & nd2name,
                  const std::string &style) {
    std::string s = "arrowhead=none ";
    s += style;
    writeDOTEdge(out, anc, des, nd2name, s, true);
}

void writeOneSideOfPathPairingToDOT(std::ostream & out,
                           const NodeWithSplits * sidePar,
                           const NodeWithSplits * sideDes,
                           const NodeWithSplits * ,//scaffAnc,
                           const NodeWithSplits * scaffDes,
                           const ToDotKey & midPointKey,
                           const NamePair & namePair,
                           NodeToDotNames & nd2name,
                           const NodeWithSplits * ,//focalNode,
                           const std::string & style,
                           const char * prefix) {
    const bool isScaffoldSide = (sideDes == scaffDes);
    //constexpr bool mergeScPhAnc = false;
    const char * ancPref = (isScaffoldSide ? "" : prefix); // = (isScaffoldSide || (mergeScPhAnc && focalNode != scaffAnc)) ...
    const char * desPref = ((isScaffoldSide || sideDes->is_tip()) ? "" : prefix);
    // If we have a tip or a node outside of this focal node, use the scaffold node for the 
    //  phylo side of the graph. Not accurate, but cuts down on the # of nodes.
    ToDotKey ancK{sidePar, ancPref};
    /*if (mergeScPhAnc && focalNode != scaffAnc) {
        ancK = ToDotKey{scaffAnc, ""};
    }*/
    ToDotKey desK{sideDes, desPref};
    if (scaffDes->is_tip()) {
        desK = ToDotKey{scaffDes, ""};
    }
    
    writeNodeDOT(out, ancK, nd2name, style, true, false, false);
    writeNodeDOT(out, desK, nd2name, style, false, false, false);
    // always write the point node along the path
    nd2name[midPointKey] = namePair;
    uncheckedWriteNodeDOT(out, midPointKey, nd2name, style, false, false, true);
    // write the edges anck -> midpoint -> desk
    writeDOTEdge(out, ancK, midPointKey, nd2name, style, true);
    writeDOTEdge(out, midPointKey, desK, nd2name, style, false);
}

void writePathPairingToDOT(std::ostream & out,
                           const NodeWithSplits *n, 
                           const PathPairingWithSplits & pp,
                           NodeToDotNames & nd2name,
                           std::set<const PathPairingWithSplits *> & pathSet,
                           const char * color,
                           const char * prefix) {
    if (contains(pathSet, &pp)) {
        return;
    }
    const std::string emptyStr;
    pathSet.insert(&pp);
    std::string bstyle = edgeStyleForColor(color);
    std::string style = bstyle;
    style.append(" style=\"dashed\"");
    // The midpoint nodes are unlabeled dots with IDs that are _addr or __addr
    const std::string pname ="_" + std::to_string(reinterpret_cast<long>(&pp));
    const std::string sname ="__" + std::to_string(reinterpret_cast<long>(&pp));
    const NamePair pv{pname, emptyStr};
    const NamePair sv{sname, emptyStr};
    const auto * pn = pp.phylo_parent;
    const auto * pd = pp.phylo_child;
    const auto * sd = pp.scaffold_des;
    const auto * sn = pp.scaffold_anc;
    const ToDotKey pk{pd, "_phpath"};
    writeOneSideOfPathPairingToDOT(out, pn, pd, sn, sd, pk, pv, nd2name, n, style, prefix);
    style = bstyle;
    style.append(" style=\"solid\"");
    const ToDotKey sk{sd, "_scpath"};
    writeOneSideOfPathPairingToDOT(out, sn, sd, sn, sd, sk, sv, nd2name, n, style, prefix);
    style = bstyle;
    style.append(" style=\"dotted\"");
    writeDOTCrossEdge(out, pk, sk, nd2name, style);
}

void writeDOTEmbeddingForNode(std::ostream & out,
                              const NodeWithSplits * nd,
                              const NodeEmbeddingWithSplits & thr,
                              const std::map<const NodeWithSplits *, NodeEmbeddingWithSplits> & eForNd,
                              NodeToDotNames & nd2name,
                              std::set<const PathPairingWithSplits *> & pathSet,
                              const char * color,
                              const char * prefix,
                              std::size_t treeIndex
                              ) {
    auto & exits = thr.get_exit_embeddings();
    const auto eait = exits.find(treeIndex);
    if (eait != exits.end()) {
        for (const auto & pp : eait->second) {
            writePathPairingToDOT(out, nd, *pp, nd2name, pathSet, color, prefix);
        }
    }
    const auto incoming = thr.get_all_incoming_path_pairs(eForNd, treeIndex);
    for (auto pp : incoming) {
        writePathPairingToDOT(out, nd, *pp, nd2name, pathSet, color, prefix);
    }
}

void write_dot_for_embedding(std::ostream & out,
                     const NodeWithSplits * nd,
                     const std::vector<TreeMappedWithSplits *> & tv,
                     const std::map<const NodeWithSplits *, NodeEmbeddingWithSplits> & eForNd,
                     bool entireSubtree,
                     bool includeLastTree) {
    NodeToDotNames nd2name;
    std::string emptyStr;
    out << "digraph G{\n";
    for (auto n : iter_pre_n_const(nd)) {
        const ToDotKey k{n, ""};
        writeNodeDOT(out, k, nd2name, "", false, true, false);
    }
    auto ndp = nd->get_parent();
    if (ndp != nullptr) {
        const ToDotKey k{ndp, ""};
        writeNodeDOT(out, k, nd2name, "", false, true, false);
    }
    for (auto n : iter_pre_n_const(nd)) {
        auto p = n->get_parent();
        if (p == nullptr) {
            continue;
        }
        const ToDotKey nk{n, ""};
        const ToDotKey pk{p, ""};
        writeDOTEdge(out, pk, nk, nd2name, emptyStr, false);
    }
    std::set<const PathPairingWithSplits *> pathSet;
    const auto nt = tv.size() - (includeLastTree ? 0U : 1U);
    for (auto n : iter_pre_n_const(nd)) {
        const auto emIt = eForNd.find(n);
        if (emIt == eForNd.end()) {
            writeDOTForNodeWithoutEmbedding(out, n, nd2name);
            continue;
        }
        const NodeEmbeddingWithSplits & thr = eForNd.at(n);
        for (auto i = 0U; i < nt; ++i) {
            const std::string tP = std::string("t") + std::to_string(i);
            auto colorIndex = std::min(LAST_COLOR_IND, i);
            const char * color = COLORS[colorIndex];
            writeDOTEmbeddingForNode(out, n, thr, eForNd, nd2name, pathSet, color, tP.c_str(), i);
        }
        if (!entireSubtree) {
            break;
        }
    }
    out << "}\n";
}


template<typename T>
using GCList = std::list<std::pair<T *, PhyloStatementSource> >;
template<typename T>
using NdToGCListMapping = std::map<T*, GCList<T> >;
template<typename T>
using NdToGCMapping = std::map<T*, std::pair<T *, PhyloStatementSource> >;
template<typename T, typename U>
void writeDOTExclusion(std::ostream & out,
                       const T * node,
                       const U & exclusion,
                       const std::string & prefix,
                       const std::string & connPrefix,
                       NodeToDotNames & nd2name,
                       const std::string & ndStyle,
                       const std::string & style) {
    for (auto n : iter_pre_n_const(node)) {
        const auto & excForThisNode = exclusion.get_nodes_excluded_from_node(n);
        const ToDotKey ancK{n, connPrefix};
        for (auto en : excForThisNode) {
            const ToDotKey desK{en, prefix};
            if ((!contains(nd2name, ancK)) || (!contains(nd2name, desK))) {
                LOG(DEBUG) << "in writeDOTExclusion edge skipped...";
                continue;
            }
            writeNodeDOT(out, desK, nd2name, ndStyle, false, true, false);
            writeDOTEdge(out, ancK, desK, nd2name, style, false);
        }
    }
}
template<typename T>
void writeDOTGroupingConstraintSingle(std::ostream & out,
                                 const NdToGCMapping<T> & ndToGCList,
                                 const std::string & prefix,
                                 const std::string & connPrefix,
                                 NodeToDotNames & nd2name,
                                 const std::string & ndStyle,
                                 const std::string & style) {
    for (const auto & incNdGCPair : ndToGCList ) {
        const auto & ind = incNdGCPair.first;
        const ToDotKey desK{incNdGCPair.first, prefix};
        writeNodeDOT(out, desK, nd2name, ndStyle, false, true, false);
        const ToDotKey ancK{incNdGCPair.second.first, connPrefix}; // use connPrefix for the prefix to connect to a connected node
        writeDOTEdge(out, ancK, desK, nd2name, style, false);
    }
}

void writeDOTForFtree(std::ostream & out,
                              const FTree<RTSplits, MappedWithSplitsData> & tree, 
                              NodeToDotNames & nd2name,
                              const char * color,
                              std::size_t treeIndex
                              ) {
    std::string esty = edgeStyleForColor(color);
    const std::string nsty = esty;
    std::string tn = "t";
    tn += std::to_string(treeIndex);
    auto r = tree.get_root();
    for (auto n : iter_pre_n_const(r)) {
        auto p = n->get_parent();
        if (p == nullptr) {
            continue;
        }
        const ToDotKey ancK{p, tn};
        ToDotKey desK{n, tn};
        if (n->is_tip()) {
            desK = ToDotKey{n, ""};
        }
        writeNodeDOT(out, ancK, nd2name, nsty, false, true, false);
        writeNodeDOT(out, desK, nd2name, nsty, false, true, false);
        writeDOTEdge(out, ancK, desK, nd2name, esty, false);
    }
    /*
    const std::string itn = std::string("inc") + tn;
    const auto & inc = tree.getIncluded2ConstraintMap();
    std::string incStyle = "shape=invtriangle color=\"";
    incStyle += color;
    incStyle += "\"";
    writeDOTGroupingConstraintSingle(out, inc, itn, tn, nd2name, incStyle, esty);
    */
    const std::string etn = std::string("exc") + tn;
    const auto & exc = tree.get_exclusions();
    std::string excStyle = "shape=octagon color=\"";
    excStyle += color;
    excStyle += "\"";
    writeDOTExclusion(out, r, exc, etn, tn, nd2name, excStyle, esty);
}

ToDotKey findAnyKey(const NodeWithSplits *nd, const NodeToDotNames & nd2name) {
    for (auto kv : nd2name) {
        if (kv.first.first == nd) {
            return kv.first;
        }
    }
    OTC_UNREACHABLE;
}
template<typename T>
void writeDOTBandsForForest(std::ostream & out, const T & bandList, NodeToDotNames & nd2name) {
    const std::string esty = "style=\"dotted\"";
    for (const auto & b : bandList) {
        if (b.is_single_tree_band()) {
            continue;
        }
        const NodeWithSplits * prev = nullptr;
        for (auto n : b.get_banded_nodes()) {
            if (prev != nullptr) {
                ToDotKey ancK = findAnyKey(prev, nd2name);
                ToDotKey desK = findAnyKey(n, nd2name);
                writeDOTEdge(out, ancK, desK, nd2name, esty, false);
            }
            prev = n;
        }
    }
}

void write_dot_forest(std::ostream & out, const RootedForest<RTSplits, MappedWithSplitsData> &forest) {
    const auto & o2n = forest.get_ott_id_to_node_mapping();
    NodeToDotNames nd2name;
    std::string emptyStr;
    out << "digraph G{\n";
    for (auto & oidNodePair : o2n) {
        auto n = oidNodePair.second;
        bool writePlainNd = forest.is_attached(n->get_ott_id());
        if (!writePlainNd && (!forest.is_in_a_band(n)) && (!forest.has_nodes_excluded_from_it(n))) {
            writePlainNd = true;
        }
        if (writePlainNd) {
            const ToDotKey k{n, ""};
            writeNodeDOT(out, k, nd2name, emptyStr, false, true, false);
        }
    }
    const auto & trees = forest.get_trees();
    auto i = 0U;
    for (const auto & tiTrPair : trees) {
        auto colorIndex = std::min(LAST_COLOR_IND, i);
        const char * color = COLORS[colorIndex];
        const auto & tree = tiTrPair.second;
        writeDOTForFtree(out, tree, nd2name, color, i);
        ++i;
    }
    writeDOTBandsForForest(out, forest.get_all_bands(), nd2name);
    out << "}\n";
}

} // namespace otc

